@@ -152,7 +152,7 @@
 				; lazily generated list of this level's threads
 
 (define (make-command-level repl-thunk repl-data dynamic-env levels throw)
-  (let ((level (really-make-command-level (make-thread-queue)
+  (let ((level (really-make-command-level (make-queue)
 					  (make-counter)
 					  dynamic-env
 					  levels
@@ -172,7 +172,7 @@
   (let ((thread (make-thread thunk (command-level-dynamic-env level) id)))
     (set-thread-scheduler! thread (command-thread))
     (set-thread-data! thread level)
-    (enqueue-thread! (command-level-queue level) thread)
+    (enqueue! (command-level-queue level) thread)
     (increment-counter! (command-level-thread-counter level))
     thread))
 
@@ -294,12 +294,7 @@
 	(*out?* #f))
     (for-each (lambda (thread)
 		(if (thread-continuation thread)
-		    (begin
-                      (remove-thread-from-queues! thread)
-		      (interrupt-thread thread
-					(lambda ignore
-					  (terminate-current-thread)))
-		      (enqueue-thread! queue thread))))
+		    (terminate-level-thread thread level)))
 	      threads)
     (dynamic-wind
      (lambda ()
@@ -313,6 +308,16 @@
 	 (if (not (null? levels))
 	     (reset-command-input! (car levels))))))))
 
+; Put the thread on the runnable queue if it is not already there and then
+; terminate it.  Termination removes the thread from any blocking queues
+; and interrupts with a throw that will run any pending dynamic-winds.
+
+(define (terminate-level-thread thread level)
+  (let ((queue (command-level-queue level)))
+    (if (not (on-queue? thread queue))
+	(enqueue! queue thread))
+    (terminate-thread! thread)))
+
 (define (reset-command-input! level)
   (let ((repl (command-level-repl-thread level)))
     (if repl
@@ -367,8 +372,7 @@
 		  (error "non-command-level thread restarted on a command level"
 			 thread))
 		 ((memq level levels)
-		  (enqueue-thread! (command-level-queue level)
-				   thread))
+		  (enqueue! (command-level-queue level) thread))
 		 (else
 		  (warn "dropping thread from exited command level"
 			thread)))
@@ -447,7 +451,7 @@
 	       (if repl-thread
 		   (begin 
 		     (set-command-level-repl-thread! level #f)
-		     (kill-thread! repl-thread)))))
+		     (terminate-level-thread repl-thread level)))))
  	    ((eq? token repl-data-token)
  	     (command-level-repl-data level))
  	    ((eq? token set-repl-data!-token)
@@ -524,14 +528,10 @@
 
 (define (kill-paused-thread! level)
   (let ((paused (command-level-paused-thread level)))
-    (if (not paused)
-	(error "level has no paused thread" level))
-    (if (eq? paused (command-level-repl-thread level))
-	(spawn-repl-thread! level))
-    (interrupt-thread paused terminate-current-thread)
-;		      (lambda ignore
-;			(terminate-current-thread)))
-    ;(enqueue-thread! (command-level-queue level) paused)
-    (set-command-level-paused-thread! level #f)))
-
-
+    (if paused
+	(begin
+	  (if (eq? paused (command-level-repl-thread level))
+	      (spawn-repl-thread! level))
+	  (terminate-thread! paused)	; it's already running, so no enqueue
+	  (set-command-level-paused-thread! level #f))
+	(warn "level has no paused thread" level))))
