@@ -176,35 +176,47 @@
 
 (define (set-port-buffering port policy . maybe-size) 
   (cond ((and (fdport? port) (open-input-port? port))
-	 (let ((size (if (pair? maybe-size) (car maybe-size) 255)))
+	 (let ((size (if (pair? maybe-size) (car maybe-size) 
+			 (if (= policy bufpol/none) 1 255))))
+	   (if (<= size 0) 
+	       (error "buffer size must be at least 1 for input ports" 
+		      port policy size))
 	   (set-input-port-buffering port policy size)))
 	((and (fdport? port) (open-output-port? port))
-	 (let ((size (if (pair? maybe-size) (car maybe-size) 255)))
-	    (if (<= size 0) (error "size must be at least 1"))
+	 (let ((size (if (pair? maybe-size) (car maybe-size) 
+			 (if (= policy bufpol/none) 0 255))))
+	   (if (< size 0) 
+	       (error "buffer size must be at least 0 for output ports" 
+		      port policy size))
 	   (set-output-port-buffering port policy size)))
 	(else
-            (warn "port-type not supported" port))))
+            (error "Not a port" port))))
 
-(define (set-output-port-buffering port policy size) 
+(define (set-output-port-buffering port policy size)
   (cond ((eq? policy bufpol/none)
+	 (if (not (= size 0))
+	     (error "buffer size must be 0 for bufpol/none on output ports" 
+		    port policy size))
 	 (install-nullbuffer port unbuffered-output-fdport-handler))
 	((eq? policy bufpol/block)
-	 (let ((old-size (byte-vector-length (port-buffer port)))
-	       (new-buffer (make-byte-vector size 0)))
-	   (if (< size old-size)
-	       (begin
-		 (really-force-output port)
-		 (obtain-port-lock port)
- 		 (set-port-index! port 0))
-	       (begin 
-		 (obtain-port-lock port)
-		 (copy-bytes! (port-buffer port) 0 new-buffer 0 old-size)))
-	   (install-buffer port new-buffer size)
-	   (release-port-lock port)))
-	((eq? policy bufpol/line)
+	 (if (= size 0)
+	     (install-nullbuffer port unbuffered-output-fdport-handler)
+	     (let ((old-size (byte-vector-length (port-buffer port)))
+		   (new-buffer (make-byte-vector size 0)))
+	       (if (< size old-size)
+		   (begin
+		     (really-force-output port)
+		     (obtain-port-lock port)
+		     (set-port-index! port 0))
+		   (begin 
+		     (obtain-port-lock port)
+		     (copy-bytes! (port-buffer port) 0 new-buffer 0 old-size)))
+	       (install-buffer port new-buffer size)
+	       (release-port-lock port))))
+	 ((eq? policy bufpol/line)
 	 ;(install-nullbuffer port (make-line-output-proc size)))
 	 (error "bufpol/line is currently not supported"))
-	(else (warn "policy not supported " policy))))
+	(else (error "policy not supported " policy))))
 
 (define (install-nullbuffer port handler)
  (really-force-output port)
@@ -253,13 +265,15 @@
 	
 (define (set-input-port-buffering port policy size)
   (cond ((eq? policy bufpol/none)
+	 (if (not (= size 1))
+	     (error "buffer size must be 1 for bufpol/none on input ports" 
+		    port policy size))
 	 (set-input-port-buffering port bufpol/block 1))
 	((eq? policy bufpol/block)
-	 (if (<= size 0) (error "size must be at least 1"))
 	 (install-input-handler port input-fdport-handler size #t))
 	((eq? policy bufpol/line)
 	 (error "bufpol/line not allowed on input"))
-	(else (warn "policy not supported " policy))))
+	(else (error "policy not supported " policy))))
 
 (define (install-input-handler port new-handler size gentle?)
   	 (obtain-port-lock port)
@@ -335,19 +349,21 @@
 
 ;;; replace rts/channel-port.scm begin
 (define (open-file fname flags . maybe-mode)
-  (with-cwd-aligned
-   (with-umask-aligned
-    (let ((fd (apply open-fdes fname flags maybe-mode))
-	  (access (bitwise-and flags open/access-mask)))
-      ((if (or (= access open/read) (= access open/read+write))
-	   make-input-fdport
-	   make-output-fdport)
-       fd 0)))))
+  (let ((fd (apply open-fdes fname flags maybe-mode))
+	(access (bitwise-and flags open/access-mask)))
+    ((if (or (= access open/read) (= access open/read+write))
+	 make-input-fdport
+	 make-output-fdport)
+     fd 0)))
 
 (define (open-input-file fname . maybe-flags)
   (let ((flags (:optional maybe-flags 0)))
     (open-file fname (deposit-bit-field flags open/access-mask open/read))))
 
+(define (deposit-bit-field bits mask field)
+  (bitwise-ior (bitwise-and field mask)
+	       (bitwise-and bits  (bitwise-not mask))))
+
 (define (open-output-file fname . rest)
   (let* ((flags (if (pair? rest) (car rest)
 		    (bitwise-ior open/create open/truncate))) ; default
@@ -469,22 +485,13 @@
   current-error-port)
 
 
-(define old-inport #f)	; Just because.
-(define old-outport #f)
-(define old-errport #f)
-
 (define (init-fdports!)
-  (if (not (fdport? (current-input-port)))
-      (set! old-inport (current-input-port)))
-  (if (not (fdport? (current-output-port)))
-      (set! old-outport (current-output-port)))
-  (if (not (fdport? (current-error-port)))
-      (set! old-errport (current-error-port)))
   (set-fluid! $current-input-port  
 	      (channel-port->input-fdport (current-input-port)))
+  (set-port-buffering (current-input-port) bufpol/none)
+
   (set-fluid! $current-output-port 
 	      (channel-port->output-fdport (current-output-port)))
-  
   (set-fluid! $current-error-port  
 	      (channel-port->unbuffered-output-fdport (current-error-port)))
   (set-fluid! $current-noise-port 
@@ -548,7 +555,7 @@
   (force-output fdport))
 
 (define (flush-all-ports)
-  (let ((thunks (output-port-forcers #f)))
+  (let ((thunks (output-port-forcers #f #f)))
     (cond ((null? thunks)
 	   #f)
 	  (else
@@ -576,6 +583,10 @@
        (thunk)))
     (placeholder-value placeholder)))
 
+(define (flush-all-ports-no-threads)
+  (let ((thunks (output-port-forcers #f #t)))
+    (for-each (lambda (thunk) (thunk)) thunks)))
+
 ;;; Extend R4RS i/o ops to handle file descriptors.
 ;;; -----------------------------------------------
 
@@ -612,7 +623,7 @@
     (let ((stream (:optional maybe-i/o (current-output-port))))
       (cond ((output-port? stream) (s48name arg ... stream))
 	    ((integer? stream) body ...)
-	    (else (error "Not a outport or file descriptor" stream))))))
+	    (else (error "Not an outport or file descriptor" stream))))))
 
 ;;; This one depends upon S48's string ports.
 (define-r4rs-output (display object) output s48-display
@@ -694,18 +705,16 @@
 
 (define (call-with-mumble-file open close)
   (lambda (string proc)
-    (with-cwd-aligned
-     (with-umask-aligned
-      (let ((port #f))
-	(dynamic-wind (lambda ()
-			(if port
-			    (warn "throwing back into a call-with-...put-file"
-				  string)
-			    (set! port (open string))))
-		      (lambda () (proc port))
-		      (lambda ()
-			(if port
-			    (close port)))))))))
+    (let ((port #f))
+      (dynamic-wind (lambda ()
+		      (if port
+			  (warn "throwing back into a call-with-...put-file"
+				string)
+			  (set! port (open string))))
+		    (lambda () (proc port))
+		    (lambda ()
+		      (if port
+			  (close port)))))))
 
 ;;; replace rts/channel-port.scm begin
 (define call-with-input-file
@@ -726,4 +735,311 @@
 
 ;;; replace rts/channel-port.scm end
 	  
-    
\ No newline at end of file
+;;; select
+;;; -----
+
+(define (port/fdes->input-port port/fd)
+  (if (port? port/fd)
+      port/fd
+      (fdes->inport port/fd)))
+
+(define (port/fdes->output-port port/fd)
+  (if (port? port/fd)
+      port/fd
+      (fdes->outport port/fd)))
+
+(define (input-port/fdes-ready? port/fd)
+  (let ((port (port/fdes->input-port port/fd)))
+    ((port-handler-ready? (port-handler port)) port)))
+
+(define (output-port/fdes-ready? port/fd)
+  (let ((port (port/fdes->output-port port/fd)))
+    ((port-handler-ready? (port-handler port)) port)))
+
+(define (make-any-ready port/fdes-ready?)
+  (lambda (port/fds)
+    (let loop ((port/fds port/fds))
+      (if (null? port/fds)
+	  '()
+	  (let ((port/fd (car port/fds)))
+	    (if (port/fdes-ready? port/fd)
+		;; one is ready, get them all
+		(let loop ((rest (cdr port/fds))
+			   (ready (list port/fd)))
+		  (cond
+		   ((null? rest) (reverse ready))
+		   ((port/fdes-ready? (car rest))
+		    (loop (cdr rest) (cons (car rest) ready)))
+		   (else
+		    (loop (cdr rest) ready))))
+		(loop (cdr port/fds))))))))
+
+(define any-input-ready (make-any-ready input-port/fdes-ready?))
+(define any-output-ready (make-any-ready output-port/fdes-ready?))
+
+(define any-channel-ready (make-any-ready fdport-channel-ready?))
+
+(define (make-port/fdes-check-unlocked input?)
+  (let ((port/fdes->port
+	 (if input?
+	     port/fdes->input-port
+	     port/fdes->output-port)))
+    (lambda (port/fd)
+      (if (port-locked? (port/fdes->port port/fd))
+	  (begin
+	    ((structure-ref interrupts enable-interrupts!))
+	    (error "SELECT on port with pending operation"
+		   port/fd))))))
+
+(define input-port/fdes-check-unlocked (make-port/fdes-check-unlocked #t))
+(define output-port/fdes-check-unlocked (make-port/fdes-check-unlocked #f))
+
+(define (port/fdes->input-channel port/fd)
+  (fdport-data:channel
+   (fdport-data
+    (port/fdes->input-port port/fd))))
+
+(define (port/fdes->output-channel port/fd)
+  (fdport-data:channel
+   (fdport-data
+    (port/fdes->output-port port/fd))))
+
+(define (select read-vec write-vec exception-vec . maybe-timeout)
+  (let ((read-list (vector->list read-vec))
+	(write-list (vector->list write-vec))
+	(timeout (:optional maybe-timeout #f)))
+
+    ((structure-ref interrupts disable-interrupts!))
+    
+    (for-each input-port/fdes-check-unlocked read-list)
+    (for-each output-port/fdes-check-unlocked write-list)
+
+    (let ((any-read (any-input-ready (filter input-port? read-list)))
+	  (any-write (any-output-ready (filter output-port? write-list))))
+
+      (if (or (eqv? timeout 0) (pair? any-read) (pair? any-write))
+	  (begin
+	    ((structure-ref interrupts enable-interrupts!))
+	    (values (list->vector any-read)
+		    (list->vector any-write)
+		    (make-vector 0)))
+	    
+	  ;; we need to block
+	  (let ((read-channels (map port/fdes->input-channel read-list))
+		(write-channels (map port/fdes->output-channel write-list)))
+
+	    (for-each (lambda (channel)
+			(add-pending-channel channel #t))
+		      read-channels)
+
+	    (for-each (lambda (channel)
+			(add-pending-channel channel #f))
+		      write-channels)
+
+	    (call-with-values
+	     (lambda ()
+	       (wait-for-channels read-channels write-channels
+				  (and timeout (* 1000 timeout))))
+	     ;; re-enables interrupts
+	     (lambda (ready-read-channels ready-write-channels)
+	       (let ((ready-read-port/fds '())
+		     (ready-write-port/fds '()))
+		 (for-each (lambda (port/fd channel)
+			     (if (memq channel ready-read-channels)
+				 (set! ready-read-port/fds
+				       (cons port/fd ready-read-port/fds))))
+			   read-list read-channels)
+		 (for-each (lambda (port/fd channel)
+			     (if (memq channel ready-write-channels)
+				 (set! ready-write-port/fds
+				       (cons port/fd ready-write-port/fds))))
+			   write-list write-channels)
+
+		 (values (list->vector (reverse ready-read-port/fds))
+			 (list->vector (reverse ready-write-port/fds))
+			 (make-vector 0))))))))))
+
+(define (select! read-vec write-vec exception-vec . maybe-timeout)
+  (let ((read-list (vector->list read-vec))
+	(write-list (vector->list write-vec))
+	(timeout (:optional maybe-timeout #f)))
+
+    ((structure-ref interrupts disable-interrupts!))
+    
+    (for-each input-port/fdes-check-unlocked read-list)
+    (for-each output-port/fdes-check-unlocked write-list)
+
+    (let ((any-read (any-input-ready (filter input-port? read-list)))
+	  (any-write (any-output-ready (filter output-port? write-list))))
+      (if (or (eqv? timeout 0) (pair? any-read) (pair? any-write))
+	  (begin
+	    ((structure-ref interrupts enable-interrupts!))
+	    (let ((n-read-ready
+		   (let ((length (vector-length read-vec)))
+		     (let loop ((i 0) (n 0))
+		       (cond
+			((= i length) n)
+			((memq (vector-ref read-vec i) any-read)
+			 (loop (+ 1 i) (+ 1 n)))
+			(else
+			 (vector-set! read-vec i #f)
+			 (loop (+ 1 i) n))))))
+		  (n-write-ready
+		   (let ((length (vector-length write-vec)))
+		     (let loop ((i 0) (n 0))
+		       (cond
+			((= i length) n)
+			((memq (vector-ref write-vec i) any-write)
+			 (loop (+ 1 i) (+ 1 n)))
+			(else
+			 (vector-set! write-vec i #f)
+			 (loop (+ 1 i) n)))))))
+		  
+	      ;; zero out EXCEPTION-VEC
+	      (let ((length (vector-length exception-vec)))
+		(let loop ((i 0))
+		  (if (< i length)
+		      (begin
+			(vector-set! exception-vec i #f)
+			(loop (+ 1 i))))))
+			  
+	      (values n-read-ready n-write-ready 0)))
+	    
+	  ;; we need to block
+	  (let ((read-channels (map port/fdes->input-channel read-list))
+		(write-channels (map port/fdes->output-channel write-list)))
+
+	    (for-each (lambda (channel)
+			(add-pending-channel channel #t))
+		      read-channels)
+
+	    (for-each (lambda (channel)
+			(add-pending-channel channel #f))
+		      write-channels)
+	
+	    (call-with-values
+	     (lambda ()
+	       (wait-for-channels read-channels write-channels
+				  (and timeout (* 1000 timeout))))
+	     ;; re-enables interrupts
+	     (lambda (ready-read-channels ready-write-channels)
+
+	       (let ((n-read-ready
+		      (let loop ((read-channels read-channels)
+				 (n-ready 0)
+				 (index 0))
+			(if (null? read-channels)
+			    n-ready
+			    (if (memq (car read-channels) ready-read-channels)
+				(loop (cdr read-channels)
+				      (+ 1 n-ready)
+				      (+ 1 index))
+				(begin
+				  (vector-set! read-vec index #f)
+				  (loop (cdr read-channels)
+					n-ready
+					(+ 1 index)))))))
+		     (n-write-ready 
+		      (let loop ((write-channels write-channels)
+				 (n-ready 0)
+				 (index 0))
+			(if (null? write-channels)
+			    n-ready
+			    (if (memq (car write-channels) ready-write-channels)
+				(loop (cdr write-channels)
+				      (+ 1 n-ready)
+				      (+ 1 index))
+				(begin
+				  (vector-set! write-vec index #f)
+				  (loop (cdr write-channels)
+					n-ready
+					(+ 1 index))))))))
+		 ;; zero out EXCEPTION-VEC
+		 (let ((length (vector-length exception-vec)))
+		   (let loop ((i 0))
+		     (if (< i length)
+			 (begin
+			   (vector-set! exception-vec i #f)
+			   (loop (+ 1 i))))))
+
+		 (values n-read-ready n-write-ready 0)))))))))
+
+(define (verify-select-port-arguments! proc lst)
+  (for-each (lambda (p)
+	      (or (input-port? p)
+		  (output-port? p)
+		  (call-error "Argument not a port" proc p)))
+	    lst))
+
+(define (select-ports timeout . ports)
+  (verify-select-port-arguments! select-ports ports)
+  (let ((read-list (filter input-port? ports))
+	(write-list (filter output-port? ports)))
+
+    ((structure-ref interrupts disable-interrupts!))
+    
+    (for-each input-port/fdes-check-unlocked read-list)
+    (for-each output-port/fdes-check-unlocked write-list)
+
+    (let ((any-read (any-input-ready read-list))
+	  (any-write (any-output-ready write-list)))
+
+      (if (or (eqv? timeout 0) (pair? any-read) (pair? any-write))
+	  (begin
+	    ((structure-ref interrupts enable-interrupts!))
+	    (append any-read any-write))
+	    
+	  (really-select-port-channels timeout read-list write-list)))))
+
+(define (select-port-channels timeout . ports)
+  (verify-select-port-arguments! select-ports ports)
+  (let ((read-list (filter input-port? ports))
+	(write-list (filter output-port? ports)))
+
+    ((structure-ref interrupts disable-interrupts!))
+    
+    (for-each input-port/fdes-check-unlocked read-list)
+    (for-each output-port/fdes-check-unlocked write-list)
+
+    (let ((any-read (any-channel-ready read-list))
+	  (any-write (any-channel-ready write-list)))
+      
+      (if (or (eqv? timeout 0) (pair? any-read) (pair? any-write))
+	  (begin
+	    ((structure-ref interrupts enable-interrupts!))
+	    (append any-read any-write))
+	  
+	  (really-select-port-channels timeout read-list write-list)))))
+
+;; assumes interrupts are disabled and that ports aren't locked
+
+(define (really-select-port-channels timeout read-list write-list)
+  (let ((read-channels (map port/fdes->input-channel read-list))
+	(write-channels (map port/fdes->output-channel write-list)))
+    
+    (for-each (lambda (channel)
+		(add-pending-channel channel #t))
+	      read-channels)
+    
+    (for-each (lambda (channel)
+		(add-pending-channel channel #f))
+	      write-channels)
+
+    (call-with-values
+     (lambda ()
+       (wait-for-channels read-channels write-channels
+			  (and timeout (* 1000 timeout))))
+     ;; re-enables interrupts
+     (lambda (ready-read-channels ready-write-channels)
+       (append (filter (lambda (read-port)
+			 (any (lambda (read-channel)
+				(eq? read-channel
+				     (port/fdes->input-channel read-port)))
+			      ready-read-channels))
+		       read-list)
+	       (filter (lambda (write-port)
+			 (any (lambda (write-channel)
+				(eq? write-channel
+				     (port/fdes->output-channel write-port)))
+			      ready-write-channels))
+		       write-list))))))
