@@ -15,22 +15,22 @@
 ;;; submatches, then POST-DSM as dead submatches.
 
 (define-record-type re-dsm :re-dsm
-  (%%make-re-dsm body pre-dsm tsm posix)
+  (really-make-re-dsm body pre-dsm tsm posix)
   re-dsm?
   (body     re-dsm:body)		; A Regexp
   (pre-dsm  re-dsm:pre-dsm)		; Integer -- initial dead submatches
   (tsm      re-dsm:tsm)					; Total submatch count
   (posix    re-dsm:posix set-re-dsm:posix))		; Posix bits
 
-(define (%make-re-dsm body pre-dsm tsm) (%%make-re-dsm body pre-dsm tsm #f))
+(define (make-re-dsm/tsm body pre-dsm tsm) (really-make-re-dsm body pre-dsm tsm #f))
 
 ;;; This is only used in code that the (RX ...) macro produces 
 ;;; for static regexps.
-(define (%make-re-dsm/posix body pre-dsm tsm posix-str tvec)
-  (%%make-re-dsm body pre-dsm tsm (new-cre posix-str tvec)))
+(define (make-re-dsm/posix body pre-dsm tsm posix-str tvec)
+  (really-make-re-dsm body pre-dsm tsm (new-cre posix-str tvec)))
 
 (define (make-re-dsm body pre-dsm post-dsm)
-  (%make-re-dsm body pre-dsm (+ post-dsm pre-dsm (re-tsm body))))
+  (make-re-dsm/tsm body pre-dsm (+ post-dsm pre-dsm (re-tsm body))))
 
 ;;; "Virtual field" for the RE-DSM record -- how many dead submatches 
 ;;; come after the body:
@@ -50,7 +50,7 @@
       (let ((pre-dsm (+ pre-dsm pre-dsm1)))
 
 	(if (= tsm (re-tsm body1)) body1		; Trivial DSM
-	    (%make-re-dsm body1 pre-dsm tsm))))))	; Non-trivial DSM
+	    (make-re-dsm/tsm body1 pre-dsm tsm))))))	; Non-trivial DSM
 
 ;;; Take a regexp RE and return an equivalent (re', pre-dsm) pair of values.
 ;;; Recurses into DSM records. It is the case that 
@@ -65,26 +65,31 @@
 
 
 
-;;; Sequence: (& re ...)
+;;; Sequence: (: re ...)
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (define-record-type re-seq :re-seq
-  (%%make-re-seq elts tsm posix)
+  (really-make-re-seq elts tsm posix)
   re-seq?
   (elts   re-seq:elts)				; Regexp list 
   (tsm    re-seq:tsm)				; Total submatch count
   (posix  re-seq:posix set-re-seq:posix))	; Posix record
 
-(define (%make-re-seq elts tsm) (%%make-re-seq elts tsm #f))
+(define (make-re-seq/tsm elts tsm) (really-make-re-seq elts tsm #f))
 
 ;;; This is only used in code that (RE ...) macro produces for static regexps.
-(define (%make-re-seq/posix elts tsm posix-str tvec)
-  (%%make-re-seq elts tsm (new-cre posix-str tvec)))
+(define (make-re-seq/posix elts tsm posix-str tvec)
+  (really-make-re-seq elts tsm (new-cre posix-str tvec)))
 
 (define (make-re-seq res)
-  (%make-re-seq res
-		(fold (lambda (re sm-count) (+ (re-tsm re) sm-count))
-		      0 res)))
+  (make-re-seq/tsm res
+		   (fold (lambda (re sm-count)
+			   (let ((maybe-tsm (re-tsm re)))
+			     (if (and (number? maybe-tsm)
+				      (number? sm-count))
+				 (+ maybe-tsm sm-count)
+				 (unspecific))))
+			 0 res)))
 
 ;;; Slightly smart sequence constructor:
 ;;; - Flattens nested sequences
@@ -98,10 +103,10 @@
 	       (if (pair? res)
 		   (let* ((re (car res))
 			  (tail (recur (cdr res))))
-		     (? ((re-seq? re)		; Flatten nested seqs
-			 (append (recur (re-seq:elts re)) tail))
-			((re-trivial? re) tail)	; Drop trivial elts
-			(else (cons re tail))))
+		     (cond ((re-seq? re) ; Flatten nested seqs
+			    (append (recur (re-seq:elts re)) tail))
+			   ((re-trivial? re) tail) ; Drop trivial elts
+			   (else (cons re tail))))
 		   '()))))
 
     (if (pair? res)
@@ -115,22 +120,29 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (define-record-type re-choice :re-choice
-  (%%make-re-choice elts tsm posix)
+  (really-make-re-choice elts tsm posix)
   re-choice?
   (elts re-choice:elts)				; List of rel-items
   (tsm  re-choice:tsm)				; Total submatch count
   (posix  re-choice:posix set-re-choice:posix))	; Posix string
 
-(define (%make-re-choice elts tsm) (%%make-re-choice elts tsm #f))
+(define (make-re-choice/tsm elts tsm) (really-make-re-choice elts tsm #f))
 
 ;;; This is only used in code that (RE ...) macro produces for static regexps.
-(define (%make-re-choice/posix elts tsm posix-str tvec)
-  (%%make-re-choice elts tsm (new-cre posix-str tvec)))
+(define (make-re-choice/posix elts tsm posix-str tvec)
+  (really-make-re-choice elts tsm (new-cre posix-str tvec)))
 
 (define (make-re-choice res)
-  (%make-re-choice res
-		   (fold (lambda (re sm-count) (+ (re-tsm re) sm-count))
-			 0 res)))
+  (if (every re-char-set? res)
+      (make-re-char-set (apply char-set-union (map re-char-set:cset res)))
+      (make-re-choice/tsm res
+                          (fold (lambda (re sm-count)
+                                  (let ((maybe-tsm (re-tsm re)))
+                                    (if (and (number? maybe-tsm)
+                                             (number? sm-count))
+                                        (+ maybe-tsm sm-count)
+                                        (unspecific))))
+                                0 res))))
 
 ;;; Slightly smart choice constructor:
 ;;; - Flattens nested choices
@@ -148,10 +160,10 @@
 	       (if (pair? res)		; & drop empty re's.
 		   (let* ((re (car res))
 			  (tail (recur (cdr res))))
-		     (? ((re-choice? re)	    ; Flatten nested choices
-			 (append (recur (re-choice:elts re)) tail))
-			((re-empty? re) tail)	    ; Drop empty re's.
-			(else (cons re tail))))
+		     (cond ((re-choice? re) ; Flatten nested choices
+			    (append (recur (re-choice:elts re)) tail))
+			   ((re-empty? re) tail) ; Drop empty re's.
+			   (else (cons re tail))))
 		   '()))))
     ;; If all elts are char-class re's, fold them together.
     (if (every static-char-class? res)
@@ -181,7 +193,7 @@
 ;;;     (= (re-repeat:tsm re) (re-tsm (re-repeat:body re)))
 
 (define-record-type re-repeat :re-repeat
-  (%%make-re-repeat from to body tsm posix)
+  (really-make-re-repeat from to body tsm posix)
   re-repeat?
   (from  re-repeat:from)	; Integer    (Macro expander abuses.)
   (to    re-repeat:to)		; Integer or #f for infinite	(Macro expander abuses.)
@@ -189,26 +201,26 @@
   (tsm   re-repeat:tsm)		; Total submatch count
   (posix re-repeat:posix set-re-repeat:posix))		; Posix record
 
-(define (%make-re-repeat from to body tsm)
-  (%%make-re-repeat from to body tsm #f))
+(define (make-re-repeat/tsm from to body tsm)
+  (really-make-re-repeat from to body tsm #f))
 
 ;;; This is only used in code that (RE ...) macro produces for static regexps.
-(define (%make-re-repeat/posix from to body tsm posix-str tvec)
-  (%%make-re-repeat from to body tsm (new-cre posix-str tvec)))
+(define (make-re-repeat/posix from to body tsm posix-str tvec)
+  (really-make-re-repeat from to body tsm (new-cre posix-str tvec)))
 
 (define (make-re-repeat from to body)
-  (%make-re-repeat  (check-arg (lambda (from)
-				 (or (not (integer? from)) ; Dynamic
-				     (>= from 0)))
-			       from
-			       make-re-repeat)
-		    (check-arg (lambda (to)
-				 (or (not (integer? to)) ; #f or dynamic
-				     (and (integer? to) (>= to 0))))
-			       to
-			       make-re-repeat)
-		    body
-		    (re-tsm body)))
+  (make-re-repeat/tsm  (check-arg (lambda (from)
+				    (or (not (integer? from)) ; Dynamic
+					(>= from 0)))
+				  from
+				  make-re-repeat)
+		       (check-arg (lambda (to)
+				    (or (not (integer? to)) ; #f or dynamic
+					(and (integer? to) (>= to 0))))
+				  to
+				  make-re-repeat)
+		       body
+		       (re-tsm body)))
 
 ;;; Slightly smart repeat constructor
 ;;; - Flattens nested repeats.
@@ -244,32 +256,33 @@
 				 dsm0)))
 		     (values from to body dsm0)))))
 
-    (? ((and (eqv? from 1) (eqv? to 1))		; re{1,1} => re
-	(values body1 pre-dsm))
+    (cond
+     ((and (eqv? from 1) (eqv? to 1))	; re{1,1} => re
+      (values body1 pre-dsm))
 
-       ((and (eqv? from 0) (eqv? to 0))		; re{0,0} => ""
-	(values re-trivial (+ (re-tsm body1) pre-dsm)))
+     ((and (eqv? from 0) (eqv? to 0))	; re{0,0} => ""
+      (values re-trivial (+ (re-tsm body1) pre-dsm)))
 
-       ;; re{m,n} => re-empty when m>n:
-       ((and (integer? from) (integer? to) (> from to))
-	(values re-empty (+ (re-tsm body1) pre-dsm)))
+     ;; re{m,n} => re-empty when m>n:
+     ((and (integer? from) (integer? to) (> from to))
+      (values re-empty (+ (re-tsm body1) pre-dsm)))
 
-       ;; Reduce the body = re-empty case.
-       ((and (re-empty? body1) (integer? from))		; (+ (in)) => (in)
-	(values (if (> from 0) re-empty re-trivial)	; (* (in)) => ""
-		pre-dsm))
+     ;; Reduce the body = re-empty case.
+     ((and (re-empty? body1) (integer? from)) ; (+ (in)) => (in)
+      (values (if (> from 0) re-empty re-trivial) ; (* (in)) => ""
+	      pre-dsm))
 
-       ;; If BODY1 is eos, bos, or "", and m<=n, reduce to simply BODY1.
-       ((and (integer? from)
-	     (or (and (integer? to) (<= from to)) (not to))
-	     (or (re-eos? body1)
-		 (re-bos? body1)
-		 (and (re-string? body1)
-		      (string=? "" (re-string:chars body1)))))
-	(values body1 pre-dsm))
+     ;; If BODY1 is eos, bos, or "", and m<=n, reduce to simply BODY1.
+     ((and (integer? from)
+	   (or (and (integer? to) (<= from to)) (not to))
+	   (or (re-eos? body1)
+	       (re-bos? body1)
+	       (and (re-string? body1)
+		    (string=? "" (re-string:chars body1)))))
+      (values body1 pre-dsm))
 
-       (else (values (make-re-repeat from to body1)	; general case
-		     pre-dsm)))))
+     (else (values (make-re-repeat from to body1) ; general case
+		   pre-dsm)))))
 
 
 
@@ -281,19 +294,19 @@
 ;;; submatches.
 
 (define-record-type re-submatch :re-submatch
-  (%%make-re-submatch body pre-dsm tsm posix)
+  (really-make-re-submatch body pre-dsm tsm posix)
   re-submatch?
   (body re-submatch:body)	; Regexp
   (pre-dsm re-submatch:pre-dsm)	; Deleted submatches preceding the body
   (tsm  re-submatch:tsm)	; Total submatch count for the record
   (posix re-submatch:posix set-re-submatch:posix)) ; Posix string
 
-(define (%make-re-submatch body pre-dsm tsm)
-  (%%make-re-submatch body pre-dsm tsm #f))
+(define (make-re-submatch/tsm body pre-dsm tsm)
+  (really-make-re-submatch body pre-dsm tsm #f))
 
 ;;; This is only used in code that (RE ...) macro produces for static regexps.
-(define (%make-re-submatch/posix body pre-dsm tsm posix-str tvec)
-  (%%make-re-submatch body pre-dsm tsm (new-cre posix-str tvec)))
+(define (make-re-submatch/posix body pre-dsm tsm posix-str tvec)
+  (really-make-re-submatch body pre-dsm tsm (new-cre posix-str tvec)))
 
 
 ;;; "Virtual field" for the RE-SUBMATCH record -- how many dead submatches 
@@ -307,7 +320,7 @@
 
 (define (make-re-submatch body . maybe-pre+post-dsm)
   (let-optionals maybe-pre+post-dsm ((pre-dsm 0) (post-dsm 0))
-    (%make-re-submatch body pre-dsm (+ pre-dsm 1 (re-tsm body) post-dsm))))
+    (make-re-submatch/tsm body pre-dsm (+ pre-dsm 1 (re-tsm body) post-dsm))))
 
 ;;; Slightly smart submatch constructor
 ;;; - DSM's unpacked
@@ -319,7 +332,7 @@
       (receive (body1 pre-dsm1) (open-dsm body)
 	(if (re-empty? body1)
 	    (re-dsm re-empty tsm 0)
-	    (%make-re-submatch body1 (+ pre-dsm pre-dsm1) tsm))))))
+	    (make-re-submatch/tsm body1 (+ pre-dsm pre-dsm1) tsm))))))
 
 
 
@@ -327,24 +340,40 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Also, re-empty and re-trivial.
 
-(define-record re-string
-  chars			; String
-  (posix #f)		; Posix record
-  ((disclose self) (list "re-string" (re-string:chars self))))
+(define-record-type re-string :re-string
+  (really-make-re-string chars posix)
+  re-string?
+  (chars re-string:chars set-re-string:chars)
+  (posix re-string:posix set-re-string:posix))
+
+(define-record-discloser :re-string
+  (lambda (r)
+    (list 're-string
+	  (re-string:chars r))))
+
+;; Kludge: POSIX wants "()" for "the empty string".
+
+(define (make-re-string chars)
+  (if (string=? "" chars)
+      re-trivial
+      (really-make-re-string chars #f)))
 
 (define re-string make-re-string)	; For consistency w/other re makers.
 
 ;;; This is only used in code that (RE ...) macro produces for static regexps.
 (define (make-re-string/posix chars posix-str tvec)
-  (let ((re (make-re-string chars)))
-    (set-re-string:posix re (new-cre posix-str tvec))
-    re))
+  (if (string=? "" chars)
+      re-trivial
+      (really-make-re-string chars (new-cre posix-str tvec))))
+
+(define re-empty-string (really-make-re-string "" #f))
 
 ;;; Matches the empty string anywhere.
-(define re-trivial (make-re-string/posix "" "" '#()))
+(define re-trivial (make-re-dsm/posix re-empty-string
+				      1 0 "()" '#()))
 
 (define (re-trivial? re)
-  (and (re-string? re) (zero? (string-length (re-string:chars re)))))
+  (eq? re re-trivial))
 
 (define-record re-char-set
   cset			; A character set	(Macro expander abuses.)
@@ -400,13 +429,16 @@
 ;;; Return the total number of submatches bound in RE.
 
 (define (re-tsm re)
-  (? ((re-seq? re)      (re-seq:tsm re))
-     ((re-choice? re)   (re-choice:tsm re))
-     ((re-repeat? re)   (re-repeat:tsm re))
-     ((re-dsm? re)      (re-dsm:tsm re))
-     ((re-submatch? re) (re-submatch:tsm re))
-     (else 0)))
-
+  (cond
+   ((re-seq? re)      (re-seq:tsm re))
+   ((re-choice? re)   (re-choice:tsm re))
+   ((re-repeat? re)   (re-repeat:tsm re))
+   ((re-dsm? re)      (re-dsm:tsm re))
+   ((re-submatch? re) (re-submatch:tsm re))
+   ((or (re-char-set? re) (re-string? re)
+	(re-bos? re) (re-eos? re)
+	(re-bol? re) (re-eol? re))
+    0)))
 
 ;;; (flush-submatches re)
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -414,19 +446,20 @@
 ;;; stripped out -- (= 0 (re-tsm (flush-submatches re))).
 
 (define (flush-submatches re)
-  (? ((zero? (re-tsm re)) re)		; RE has no submatches.
+  (cond
+   ((zero? (re-tsm re)) re)		; RE has no submatches.
 
-     ((re-seq?    re) (re-seq    (map flush-submatches (re-seq:elts    re))))
-     ((re-choice? re) (re-choice (map flush-submatches (re-choice:elts re))))
+   ((re-seq?    re) (re-seq    (map flush-submatches (re-seq:elts    re))))
+   ((re-choice? re) (re-choice (map flush-submatches (re-choice:elts re))))
 
-     ((re-repeat? re) (re-repeat (re-repeat:from re)
-				 (re-repeat:to re)
-				 (flush-submatches (re-repeat:body re))))
+   ((re-repeat? re) (re-repeat (re-repeat:from re)
+			       (re-repeat:to re)
+			       (flush-submatches (re-repeat:body re))))
 		  
-     ((re-submatch? re) (flush-submatches (re-submatch:body re)))
-     ((re-dsm? re)      (flush-submatches (re-dsm:body      re)))
+   ((re-submatch? re) (flush-submatches (re-submatch:body re)))
+   ((re-dsm? re)      (flush-submatches (re-dsm:body      re)))
 
-     (else re)))
+   (else re)))
 
 
 ;;; Map F over ELTS. (F x) returns two values -- the "real" return value,
@@ -451,70 +484,72 @@
 
 (define (uncase re)
   (receive (new-re changed?)
-           (let recur ((re re))
-	     (? ((re-seq? re)
-		 (let ((elts (re-seq:elts re)))
-		   (receive (new-elts elts-changed?)
-		            (map/changed recur elts)
-		     (if elts-changed?
-			 (values (%make-re-seq new-elts (re-seq:tsm re)) #t)
-			 (values re #f)))))
+      (let recur ((re re))
+	(cond
+	 ((re-seq? re)
+	  (let ((elts (re-seq:elts re)))
+	    (receive (new-elts elts-changed?)
+		(map/changed recur elts)
+	      (if elts-changed?
+		  (values (make-re-seq/tsm new-elts (re-seq:tsm re)) #t)
+		  (values re #f)))))
 
-		((re-choice? re)
-		 (let ((elts (re-choice:elts re)))
-		   (receive (new-elts elts-changed?)
-			    (map/changed recur elts)
-		     (if elts-changed?
-			 (values (re-choice new-elts) #t)
-			 (values re #f)))))
+	 ((re-choice? re)
+	  (let ((elts (re-choice:elts re)))
+	    (receive (new-elts elts-changed?)
+		(map/changed recur elts)
+	      (if elts-changed?
+		  (values (re-choice new-elts) #t)
+		  (values re #f)))))
 
-		  ((re-char-set? re)
-		   (let* ((cs (re-char-set:cset re))
-			  (new-cs (uncase-char-set cs))) ; Better not be code.
-		     (if (char-set= cs new-cs)
-			 (values re #f)
-			 (values (make-re-char-set new-cs) #t))))
+	 ((re-char-set? re)
+	  (let* ((cs (re-char-set:cset re))
+		 (new-cs-re (uncase-char-set cs))) ; Better not be code.
+	    (if (char-set= cs  (re-char-set:cset new-cs-re))
+		(values re #f)
+		(values new-cs-re #t))))
 
-		  ((re-repeat? re)
-		   (receive (new-body body-changed?) (recur (re-repeat:body re))
-		     (if body-changed?
-			 (values (re-repeat (re-repeat:from re)
-					    (re-repeat:to re)
-					    new-body)
-				 #t)
-			 (values re #f))))
+	 ((re-repeat? re)
+	  (receive (new-body body-changed?) (recur (re-repeat:body re))
+	    (if body-changed?
+		(values (re-repeat (re-repeat:from re)
+				   (re-repeat:to re)
+				   new-body)
+			#t)
+		(values re #f))))
 
-		  ((re-submatch? re)
-		   (receive (new-body body-changed?) (recur (re-submatch:body re))
-		     (if body-changed?
-			 (values (%make-re-submatch new-body
-						    (re-submatch:pre-dsm re)
-						    (re-submatch:tsm     re))
-				 #t)
-			 (values re #f))))
+	 ((re-submatch? re)
+	  (receive (new-body body-changed?) (recur (re-submatch:body re))
+	    (if body-changed?
+		(values (make-re-submatch/tsm new-body
+					      (re-submatch:pre-dsm re)
+					      (re-submatch:tsm     re))
+			#t)
+		(values re #f))))
 		  
-		  ((re-string? re)
-		   (let ((cf-re (uncase-string (re-string:chars re))))
-		     (if (re-string? cf-re)
-			 (values re    #f)
-			 (values cf-re #t))))
+	 ((re-string? re)
+	  (let ((cf-re (uncase-string (re-string:chars re))))
+	    (if (re-string? cf-re)
+		(values re    #f)
+		(values cf-re #t))))
 
-		  (else (values re #f))))
+	 (else (values re #f))))
     new-re))
 		  
      
 ;;; (uncase-char-set cs)
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-;;; Return a char-set cs' such that cs' contains every char c in cs in both
-;;; its upcase and downcase form.
+;;; Return a regexp for char-set cs' such that cs' contains every char
+;;; c in cs in both its upcase and downcase form.
 
 (define (uncase-char-set cs)
-  (char-set-fold (lambda (c new-cset)
-		   (char-set-adjoin! new-cset
-				     (char-downcase c)
-				     (char-upcase c)))
-		 (char-set-copy char-set:empty)
-		 cs))
+  (make-re-char-set
+   (char-set-fold (lambda (c new-cset)
+                    (char-set-adjoin! new-cset
+                                      (char-downcase c)
+                                      (char-upcase c)))
+                  (char-set-copy char-set:empty)
+                  cs)))
 
 
 ;;; I actually make an effort to keep this a re-string
@@ -525,11 +560,14 @@
 (define (uncase-string s)
   ;; SEQ is a list of chars and doubleton char-sets.
   (let* ((seq (string-fold-right (lambda (c lis)
-				   (cons (? ((char-lower-case? c) (char-set c (char-upcase   c)))
-					    ((char-upper-case? c) (char-set c (char-downcase c)))
-					    (else c))
+				   (cons (cond
+					  ((char-lower-case? c)
+					   (char-set c (char-upcase   c)))
+					  ((char-upper-case? c)
+					   (char-set c (char-downcase c)))
+					  (else c))
 					 lis))
-			    '() s))
+				 '() s))
 
 	 ;; Coalesce adjacent chars together into a string.
 	 (fixup (lambda (chars seq)
